-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THMainMenu = {}
local THMainMenu_mt = THUtils.createClass(THMainMenu, TabbedMenu)
THMainMenu.CONTROLS = {
    TITLE_TEXT = "menuTitleText",
    VERSION_TEXT = "menuVersionText"
}
THMainMenu.L10N_SYMBOL = {
    BUTTON_BACK = "button_back"
}
THMainMenuData = {}
local THMainMenuData_mt = THUtils.createClass(THMainMenuData)
local debugFlagId = THDebugUtil.createDebugFlag("THMainMenu")
THMainMenuData.debugFlagId = debugFlagId
function THMainMenu.new(target, customMt, messageCenter, i18n, inputManager)
    customMt = customMt or THMainMenu_mt
    local self = TabbedMenu.new(target, customMt, messageCenter, i18n, inputManager)
    self:registerControls(THMainMenu.CONTROLS)
    self.defaultMenuButtonInfo = {}
    self.performBackgroundBlur = true
    self.thCustomData = THMainMenuData.new(self)
    return self
end
function THMainMenu:onGuiSetupFinished(...)
    THMainMenu:superClass().onGuiSetupFinished(self, ...)
    local modData = g_thAdvancedAddon.modData
    self.menuTitleText:setText(modData.title)
    self.menuVersionText:setText(modData.version.text)
end
function THMainMenu:copyAttributes(src, ...)
    THMainMenu:superClass().copyAttributes(self, src, ...)
end
function THMainMenu:setupMenuButtonInfo(...)
    THMainMenu:superClass().setupMenuButtonInfo(self, ...)
    local onButtonBackFunction = self.clickBackCallback
    self.backButtonInfo = {
        inputAction = InputAction.MENU_BACK,
        text = self.l10n:getText(THMainMenu.L10N_SYMBOL.BUTTON_BACK),
        callback = onButtonBackFunction
    }
    self.defaultMenuButtonInfo = {
        self.backButtonInfo
    }
    self.defaultMenuButtonInfoByActions = {
        [InputAction.MENU_BACK] = self.defaultMenuButtonInfo[1]
    }
    self.defaultButtonActionCallbacks = {
        [InputAction.MENU_BACK] = onButtonBackFunction
    }
end
function THMainMenuData.new(parent, customMt, ...)
    customMt = customMt or THMainMenuData_mt
    local self = setmetatable({}, customMt)
    self.parent = parent
    self.isPagingElementInitialized = false
    return self
end
function THMainMenuData:getIsMenuValid()
    return true
end
function THMainMenuData:addMenuPageFromXML(refXMLFilename, pageName, predicate, rootEnv, ...)
    local mainMenu = self.parent
    if THUtils.argIsValid(type(refXMLFilename) == "string", "refXMLFilename", refXMLFilename)
        and THUtils.argIsValid(type(pageName) == "string" and pageName ~= "", "pageName", pageName)
        and THUtils.argIsValid(predicate == nil or type(predicate) == "function", "predicate", predicate)
        and THUtils.argIsValid(THUtils.getIsType(rootEnv, THAbstractSystem), "rootEnv", rootEnv)
    then
        local gui = rootEnv.gui
        local i18n = rootEnv.i18n
        local modName = rootEnv.modName
        local modPath = rootEnv.modPath
        local modEnvironment = THUtils.getModEnvironment(modName)
        if modEnvironment == nil then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "rootEnv", rootEnv)
            return
        end
        refXMLFilename = Utils.getFilename(refXMLFilename, modPath)
        local xmlFile = XMLFile.loadIfExists("THFrameRefTempXML", refXMLFilename)
        if xmlFile == nil then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "refXMLFilename", refXMLFilename)
        else
            local xmlHandle = xmlFile.handle
            local xmlBaseKey = string.format("thFrameReferences." .. pageName)
            local guiElementKey = string.format("%s.GuiElement(0)", xmlBaseKey)
            if not xmlFile:hasProperty(xmlBaseKey) then
                return
            end
            local menuPage = nil
            local pageXMLFilename = xmlFile:getString(xmlBaseKey .. "#filename")
            local pageClassName = xmlFile:getString(xmlBaseKey .. "#class")
            local pageClass = modEnvironment[pageClassName]
            local pageId = xmlFile:getString(xmlBaseKey .. "#id")
            if pageXMLFilename ~= nil then
                pageXMLFilename = Utils.getFilename(pageXMLFilename, modPath)
            end
            local iconFilename = xmlFile:getString(xmlBaseKey .. "#iconFilename")
            local iconUVs = xmlFile:getVector(xmlBaseKey .. "#iconUVs", nil, 4)
            if iconFilename ~= nil then
                iconFilename = Utils.getFilename(iconFilename, modPath)
            end
            if pageXMLFilename == nil or not fileExists(pageXMLFilename) then
                THUtils.xmlMsg(xmlBaseKey, "filename", pageXMLFilename, true, THXMLMessage.INVALID_VALUE)
            elseif type(pageClass) ~= "table" or type(pageClass.new) ~= "function" then
                THUtils.xmlMsg(xmlBaseKey, "class", pageClass, true, THXMLMessage.INVALID_VALUE)
            elseif pageId == nil or pageId == "" then
                THUtils.xmlMsg(xmlBaseKey, "id", pageId, true, THXMLMessage.INVALID_VALUE)
            elseif mainMenu[pageId] ~= nil then
                THUtils.xmlMsg(xmlBaseKey, "id", pageId, true, THXMLMessage.DUPLICATE_ENTRY)
            elseif iconFilename == nil or not fileExists(iconFilename) then
                THUtils.xmlMsg(xmlBaseKey, "iconFilename", iconFilename, true, THXMLMessage.INVALID_VALUE)
            elseif type(iconUVs) ~= "table" or #iconUVs ~= 4
                or type(iconUVs[1]) ~= "number" or iconUVs[1] < 0
                or type(iconUVs[2]) ~= "number" or iconUVs[2] < 0
                or type(iconUVs[3]) ~= "number" or iconUVs[3] < 0
                or type(iconUVs[4]) ~= "number" or iconUVs[4] < 0
            then
                THUtils.xmlMsg(xmlBaseKey, "iconUVs", iconUVs, true, THXMLMessage.INVALID_VALUE)
            else
                if not self.isPagingElementInitialized then
                    local function hook_addPage(pSuperFunc, pSelf, pElementId, pElement, pTitle, pIndex, ...)
                        if pSelf == mainMenu.pagingElement
                            and type(pElement) == "table" and type(pElement.name) == "string"
                            and pTitle == ""
                        then
                            THUtils.call(function()
                                if i18n:hasText("ui_" .. pElement.name) then
                                    pElementId = pElement.name:upper()
                                    pTitle = i18n:getText("ui_" .. pElement.name)
                                end
                            end)
                        end
                        return pSuperFunc(pSelf, pElementId, pElement, pTitle, pIndex, ...)
                    end
                    THUtils.setFunctionHook(mainMenu.pagingElement, "addPage", nil, hook_addPage, false)
                    self.isPagingElementInitialized = true
                end
                local function createMenuPage(...)
                    local rootElement = gui.frames[pageName]
                    local pageController = nil -- placeholder
                    if rootElement == nil then
                        setfenv(pageClass.new, modEnvironment)
                        pageController = pageClass.new(nil, nil, rootEnv, ...)
                        if pageController ~= nil then
                            gui:loadGui(pageXMLFilename, pageName, pageController, true)
                            rootElement = gui.frames[pageName]
                            if rootElement == nil then
                                THUtils.stackMsg("Could not create gui for page %q", pageName)
                                return
                            end
                        end
                    else
                        pageController = rootElement.target
                    end
                    if pageController == nil then
                        THUtils.stackMsg("Could not find page controller for page %q", pageName)
                        return
                    end
                    xmlFile:setString(guiElementKey .. "#type", "frameReference")
                    xmlFile:setString(guiElementKey .. "#name", pageName)
                    xmlFile:setString(guiElementKey .. "#ref", pageName)
                    xmlFile:setString(guiElementKey .. "#id", pageId)
                    gui:loadGuiRec(xmlHandle, xmlBaseKey, mainMenu.pagingElement, pageController)
                    mainMenu.controlIDs[pageId] = nil
                    mainMenu:registerControls({ pageId })
                    mainMenu:exposeControlsAsFields(pageId)
                    mainMenu.pagingElement:updateAbsolutePosition()
                    mainMenu.pagingElement:updatePageMapping()
                    pageController = mainMenu[pageId]
                    if pageController == nil then
                        THUtils.stackMsg("Could not add page controller for page %q to main menu", pageName)
                    else
                        if predicate == nil then
                            if pageController.makeSelfEnabledPredicate ~= nil then
                                predicate = pageController:makeSelfEnabledPredicate()
                            else
                                predicate = function()
                                    return true
                                end
                            end
                        end
                        local normalizedIconUVs = GuiUtils.getUVs(iconUVs)
                        mainMenu:registerPage(pageController, nil, predicate)
                        mainMenu:addPageTab(pageController, iconFilename, normalizedIconUVs)
                        if pageController.menuTitleText ~= nil then
                            local titleText = i18n:getText("ui_" .. pageController.name, modName)
                            pageController.menuTitleText:setText(titleText)
                        end
                        if pageController.menuFrameIcon ~= nil then
                            pageController.menuFrameIcon:setImageFilename(iconFilename)
                            pageController.menuFrameIcon:setImageUVs(nil, THUtils.unpack(normalizedIconUVs))
                        end
                        menuPage = rootEnv:getDataTable(pageController)
                        if menuPage == nil then
                            menuPage = pageController
                        end
                        if THDebugUtil.getIsEnabled(debugFlagId) then
                            THDebugUtil.printMsg(debugFlagId, nil, "Added page %q to main menu:", pageName)
                            THDebugUtil.printTable(debugFlagId, nil, menuPage, 1)
                            if menuPage ~= pageController then
                                THDebugUtil.printMsg(debugFlagId, nil, "Page controller:")
                                THDebugUtil.printTable(debugFlagId, nil, pageController, 1)
                            end
                        end
                        if menuPage.initialize ~= nil then
                            setfenv(menuPage.initialize, modEnvironment)
                            menuPage:initialize()
                        end
                    end
                end
                THUtils.call(createMenuPage, ...)
            end
            xmlFile:delete()
            return menuPage
        end
    end
end