-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THSettingsMenuFrame = {}
local THSettingsMenuFrame_mt = THUtils.createClass(THSettingsMenuFrame, TabbedMenuFrameElement)
THSettingsMenuFrame.CONTROLS = {
    MENU_FRAME_ICON = "menuFrameIcon",
    MENU_TITLE_TEXT = "menuTitleText",
    SETTINGS_CONTAINER = "settingsContainer",
    BOX_LAYOUT = "boxLayout",
    CATEGORY_TEMPLATE = "categoryTemplate",
    CHECK_BOX_TEMPLATE = "checkBoxTemplate",
    MULTI_OPTION_TEMPLATE = "multiOptionTemplate"
}
local function getCustomData(self)
    local customData = nil
    if self ~= nil then
        customData = rawget(self, self)
    end
    return customData
end
THSettingsMenuFrame.getCustomData = getCustomData
function THSettingsMenuFrame.new(target, customMt, manager)
    customMt = customMt or THSettingsMenuFrame_mt
    local self = TabbedMenuFrameElement.new(target, customMt)
    self:registerControls(THSettingsMenuFrame.CONTROLS)
    local customData = {}
    customData.manager = manager
    customData.categoryElementMapping = {}
    customData.settingElementMapping = {}
    customData.unlinkedElements = {}
    customData.areSettingsDirty = false
    customData.isFrameInitialized = false
    rawset(self, self, customData)
    return self
end
function THSettingsMenuFrame:copyAttributes(src, ...)
    THSettingsMenuFrame:superClass().copyAttributes(self, src, ...)
    local customData = getCustomData(self)
    local srcData = getCustomData(src)
    customData.manager = srcData.manager
end
function THSettingsMenuFrame:onGuiSetupFinished(...)
    THSettingsMenuFrame:superClass().onGuiSetupFinished(self, ...)
end
function THSettingsMenuFrame:onFrameOpen(...)
    THSettingsMenuFrame:superClass().onFrameOpen(self, ...)
    self:reloadSettingElements()
    self:updateSettingElements()
end
function THSettingsMenuFrame:onFrameClose(...)
    THSettingsMenuFrame:superClass().onFrameClose(self, ...)
    self:saveSettings()
end
function THSettingsMenuFrame:delete(...)
    THSettingsMenuFrame:superClass().delete(self, ...)
    local customData = getCustomData(self)
    for unlinkedElement in pairs(customData.unlinkedElements) do
        unlinkedElement:delete()
    end
end
function THSettingsMenuFrame:onMasterUserAdded(user)
    self:updateSettingElements()
end
function THSettingsMenuFrame:onSettingsChanged(manager)
    local customData = getCustomData(self)
    if customData.manager ~= nil and manager == customData.manager then
        self:updateSettingElements()
    end
end
function THSettingsMenuFrame:onClickMultiOption(state, element, ...)
    self:updateSettingElements()
end
function THSettingsMenuFrame:makeSelfEnabledPredicate()
    return function()
        local customData = getCustomData(self)
        if customData.manager ~= nil then
            local _, numSettings = customData.manager:getSettingsList()
            if numSettings > 0 then
                return true
            end
        end
        return true
    end
end
function THSettingsMenuFrame:initialize(...)
    local customData = getCustomData(self)
    if not customData.isFrameInitialized and customData.manager ~= nil then
        local manager = customData.manager
        local boxLayoutElement = self.boxLayout
        local oldInvalidateLayoutFunc = boxLayoutElement.invalidateLayout
        boxLayoutElement.invalidateLayout = function(pSelf, pIgnoreVisibility, ...)
            local function appendFunc2(...)
                THUtils.call(function()
                    local elementsArray = pSelf.elements
                    local numElements = #elementsArray
                    local focusedElement = FocusManager:getFocusedElement()
                    local firstElement, lastElement = nil, nil -- placeholders
                    pSelf.wrapAround = true
                    if numElements > 0 then
                        for elementIndex = 1, numElements do
                            local element = elementsArray[elementIndex]
                            if element.handleFocus and element:getIsVisible() and not element:getIsDisabled() then
                                if firstElement == nil then
                                    firstElement = element
                                elseif lastElement ~= nil then
                                    FocusManager:linkElements(element, FocusManager.TOP, lastElement)
                                    FocusManager:linkElements(lastElement, FocusManager.BOTTOM, element)
                                    FocusManager:linkElements(element, FocusManager.LEFT, lastElement)
                                    FocusManager:linkElements(lastElement, FocusManager.RIGHT, element)
                                end
                                lastElement = element
                            else
                                FocusManager:removeElement(element)
                                if element == focusedElement then
                                    focusedElement = nil
                                end
                            end
                        end
                        if firstElement ~= nil then
                            FocusManager:linkElements(firstElement, FocusManager.TOP, lastElement)
                            FocusManager:linkElements(lastElement, FocusManager.BOTTOM, firstElement)
                            FocusManager:linkElements(firstElement, FocusManager.LEFT, lastElement)
                            FocusManager:linkElements(lastElement, FocusManager.RIGHT, firstElement)
                        end
                    end
                    if focusedElement == nil then
                        pSelf.target:setSoundSuppressed(true)
                        if firstElement ~= nil then
                            pSelf:scrollToMakeElementVisible(firstElement)
                            FocusManager:setFocus(firstElement)
                        else
                            FocusManager:setFocus(pSelf)
                        end
                        pSelf.target:setSoundSuppressed(false)
                    end
                end)
                return ...
            end
            return appendFunc2(oldInvalidateLayoutFunc(pSelf, pIgnoreVisibility, ...))
        end
        manager.messageCenter:subscribe(MessageType.MASTERUSER_ADDED, self.onMasterUserAdded, self)
        manager.messageCenter:subscribe(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, self.onSettingsChanged, self)
        customData.isFrameInitialized = true
    end
end
function THSettingsMenuFrame:getBoxLayoutElementIndex(element)
    for otherIndex, otherElement in pairs(self.boxLayout.elements) do
        if otherElement == element then
            return otherIndex
        end
    end
    return 0
end
function THSettingsMenuFrame:getSettingCategoryElement(category)
    local customData = getCustomData(self)
    if customData.manager ~= nil then
        local manager = customData.manager
        local categoryData = manager:getSettingCategory(category)
        if categoryData ~= nil then
            for otherIndex, otherElement in pairs(self.boxLayout.elements) do
                local elementData = manager:getDataTable(otherElement)
                if elementData ~= nil and elementData.categoryIndex == categoryData.index then
                    return elementData, otherIndex
                end
            end
        end
    end
    return nil, 0
end
function THSettingsMenuFrame:getSettingElement(setting)
    local customData = getCustomData(self)
    if customData.manager ~= nil then
        local manager = customData.manager
        local settingData = manager:getSetting(setting)
        if settingData ~= nil then
            for otherIndex, otherElement in pairs(self.boxLayout.elements) do
                local elementData = manager:getDataTable(otherElement)
                if elementData ~= nil and elementData.settingIndex == settingData.index then
                    return elementData, otherIndex
                end
            end
        end
    end
    return nil, 0
end
function THSettingsMenuFrame:reloadSettingElements()
    local customData = getCustomData(self)
    if customData.manager ~= nil then
        local manager = customData.manager
        local boxLayoutElement = self.boxLayout
        local elementsArray = boxLayoutElement.elements
        local focusedElement = FocusManager:getFocusedElement()
        local firstElement = nil -- placeholder
        if #elementsArray > 0 then
            for elementIndex = #elementsArray, 1, -1 do
                local element = elementsArray[elementIndex]
                element:unlinkElement()
                FocusManager:removeElement(element)
                if customData.unlinkedElements[element] == nil then
                    customData.unlinkedElements[element] = element
                end
            end
        end
        local categoriesArray, numCategories = manager:getSettingCategoryList(true)
        if numCategories > 0 then
            for categoryIndex = 1, numCategories do
                local categoryData = categoriesArray[categoryIndex]
                local numSettings = #categoryData.settings
                local hasVisibleSetting = false
                local categoryElementData, categoryElement = manager:getDataTable(customData.categoryElementMapping[categoryData.index])
                if categoryElement == nil then
                    categoryElement = self.categoryTemplate:clone(boxLayoutElement)
                    categoryElement:setText(categoryData.title)
                    customData.categoryElementMapping[categoryData.index] = categoryElement
                else
                    boxLayoutElement:addElement(categoryElement)
                    customData.unlinkedElements[categoryElement] = nil
                end
                if categoryElementData == nil then
                    categoryElementData = manager:createDataTable(categoryElement)
                    categoryElementData.categoryIndex = categoryData.index
                end
                if numSettings > 0 then
                    for i = 1, numSettings do
                        local settingIndex = categoryData.settings[i]
                        local settingData = manager:getSetting(settingIndex)
                        local settingElementData, settingElement = manager:getDataTable(customData.settingElementMapping[settingData.index])
                        if settingElement == nil then
                            if settingData.defaultValue <= 0 then
                                THUtils.warningMsg("Setting (%s): No default value specified, skipping...")
                            elseif settingData.elementType == THSettingsManager.ELEMENT_TYPE.CHECKED then
                                settingElement = self.checkBoxTemplate:clone(boxLayoutElement)
                            elseif settingData.elementType == THSettingsManager.ELEMENT_TYPE.MULTI_OPTION then
                                settingElement = self.multiOptionTemplate:clone(boxLayoutElement)
                            end
                            if settingElement ~= nil then
                                settingElement:setTexts(settingData.valueTexts)
                                customData.settingElementMapping[settingData.index] = settingElement
                            end
                        else
                            boxLayoutElement:addElement(settingElement)
                            customData.unlinkedElements[settingElement] = nil
                        end
                        if settingElement ~= nil then
                            local changedValue, valueIndex = settingData:getChangedValue()
                            if changedValue ~= nil and valueIndex > 0 then
                                settingElement:setState(valueIndex)
                            end
                            if settingElementData == nil then
                                settingElementData = manager:createDataTable(settingElement)
                                local titleElement = settingElement:getDescendantByName("titleText")
                                local toolTipElement = settingElement:getDescendantByName("toolTipText")
                                local restartIconElement = settingElement:getDescendantByName("restartIcon")
                                titleElement:setText(settingData.title)
                                toolTipElement:setText(settingData.toolTip)
                                restartIconElement:setVisible(false)
                                settingElementData.settingIndex = settingData.index
                                settingElementData.titleElement = titleElement
                                settingElementData.toolTipElement = toolTipElement
                                settingElementData.restartIconElement = restartIconElement
                            end
                            local isSettingVisible = settingData.showInMenu
                            if isSettingVisible then
                                settingElement:setVisible(true)
                                settingElement:setDisabled(false)
                                settingElement:reloadFocusHandling(true)
                                if firstElement == nil or settingElement == focusedElement then
                                    firstElement = settingElement
                                end
                                hasVisibleSetting = true
                            else
                                settingElement:setVisible(false)
                                settingElement:setDisabled(true)
                            end
                        end
                    end
                end
                local isCategoryVisible = categoryData.showInMenu and hasVisibleSetting
                if isCategoryVisible then
                    categoryElement:setVisible(true)
                    categoryElement:setDisabled(false)
                    categoryElement:reloadFocusHandling(true)
                else
                    categoryElement:setVisible(false)
                    categoryElement:setDisabled(true)
                end
            end
        end
        if firstElement ~= nil then
            self:setSoundSuppressed(true)
            boxLayoutElement:scrollToMakeElementVisible(firstElement)
            FocusManager:setFocus(firstElement)
            self:setSoundSuppressed(false)
        end
        boxLayoutElement:invalidateLayout()
    end
end
function THSettingsMenuFrame:updateSettingElements()
    local customData = getCustomData(self)
    if customData.manager ~= nil then
        local manager = customData.manager
        local boxLayoutElement = self.boxLayout
        local elementsArray = boxLayoutElement.elements
        local numElements = #elementsArray
        local isLayoutDirty = false
        local focusedElement = FocusManager:getFocusedElement()
        local firstElement = nil -- placeholder
        if numElements > 0 then
            for elementIndex = 1, numElements do
                local elementData, element = manager:getDataTable(elementsArray[elementIndex])
                if elementData ~= nil and elementData.settingIndex ~= nil then
                    local settingData = manager:getSetting(elementData.settingIndex)
                    local isSettingDisabled = true
                    local isSettingSaved = true
                    local isRestartIconVisible = false
                    if settingData ~= nil and element:getIsVisible() then
                        local settingValue, valueIndex = settingData:getChangedValue()
                        local elementState = element:getState()
                        if settingValue ~= nil and valueIndex ~= nil
                            and elementState ~= nil and elementState > 0
                        then
                            if settingData.restartRequired then
                                if elementState ~= settingData.currentValue then
                                    isRestartIconVisible = true
                                end
                            end
                            if elementState ~= settingData.savedValue then
                                isSettingSaved = false
                            end
                            if firstElement == nil then
                                firstElement = element
                            end
                            if settingData:getCanAccess() then
                                isSettingDisabled = false
                            end
                        end
                    end
                    local titleElement = elementData.titleElement
                    local restartIconElement = elementData.restartIconElement
                    if isSettingSaved then
                        titleElement:applyProfile("multiTextOptionSettingsTitle")
                    else
                        titleElement:applyProfile("thMultiTextOptionTitleChanged")
                    end
                    restartIconElement:setVisible(isRestartIconVisible)
                    if element:getIsDisabled() ~= isSettingDisabled then
                        element:setDisabled(isSettingDisabled)
                        if element == focusedElement then
                            focusedElement = nil
                        end
                        isLayoutDirty = true
                    end
                end
            end
        end
        if isLayoutDirty then
            if focusedElement == nil and firstElement ~= nil then
                self:setSoundSuppressed(true)
                boxLayoutElement:scrollToMakeElementVisible(firstElement)
                FocusManager:setFocus(firstElement)
                self:setSoundSuppressed(false)
            end
            boxLayoutElement:invalidateLayout()
        end
    end
end
function THSettingsMenuFrame:saveSettings()
    local customData = getCustomData(self)
    if customData.manager ~= nil then
        local manager = customData.manager
        local elementsArray = self.boxLayout.elements
        local numElements = #elementsArray
        if numElements > 0 then
            for elementIndex = 1, numElements do
                local elementData, element = manager:getDataTable(elementsArray[elementIndex])
                if elementData ~= nil and elementData.settingIndex ~= nil
                    and element:getIsVisible() and not element:getIsDisabled()
                then
                    local settingData = manager:getSetting(elementData.settingIndex)
                    local elementState = element:getState()
                    if settingData ~= nil and elementState > 0 then
                        settingData:setValueByIndex(elementState)
                    end
                end
            end
        end
    end
end