-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THSettingsManager = {}
local THSettingsManager_mt = THUtils.createClass(THSettingsManager, THAbstractSystem)
THSettingsManager.SETTING_NUM_BITS = 16
THSettingsManager.SETTING_VALUE_NUM_BITS = 32
THSettingsManager.ELEMENT_TYPE = {
    CHECKED = 1,
    MULTI_OPTION = 2
}
THSettingsManager.ELEMENT_TYPE_ID = {}
for typeId, typeIndex in pairs(THSettingsManager.ELEMENT_TYPE) do
    THSettingsManager.ELEMENT_TYPE_ID[typeIndex] = typeId
end
THSettingsManager.MESSAGE_TYPE = {
    SETTINGS_CHANGED = nextMessageTypeId()
}
THSetting = {}
local THSetting_mt = THUtils.createClass(THSetting)
function THSettingsManager.initXMLSchema(xmlSchema, xmlPath)
    local settingPath = xmlPath .. ".settings.setting(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, settingPath, "#name", "Setting name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, settingPath, "#valueIndex", "Currently selected value index")
end
function THSettingsManager.initXMLSchemaSettings(xmlSchema, xmlPath)
    THSettingsManager.initXMLSchema(xmlSchema, xmlPath)
end
function THSettingsManager.initXMLSchemaSavegame(xmlSchema, xmlPath)
    THSettingsManager.initXMLSchema(xmlSchema, xmlPath)
end
function THSettingsManager.new(rootEnv, dataKey, modName, customMt)
    customMt = customMt or THSettingsManager_mt
    local self = THAbstractSystem.new(rootEnv, dataKey, modName, customMt)
    if self ~= nil then
        self.settingCategories = {
            byId = {},
            byIndex = {}
        }
        self.SETTING_CATEGORY = {}
        self.settings = {
            byId = {},
            byIndex = {}
        }
        self.SETTING = {}
        self.isClientDirty = false
        self.areSettingsLoaded = false
        self.dirtyFlag = rootEnv:getNextDirtyFlag()
        self:addSettingCategory("main", nil, nil, false)
        self.thEventManager:addEventListener(self, self.modName)
    end
    return self
end
function THSettingsManager:loadSettingsFromXMLFile(xmlFile, xmlKey)
    local loadSuccess = true
    xmlFile:iterate(xmlKey .. ".settings.setting", function(_, pSettingKey)
        local settingName = THUtils.getXMLValue(xmlFile, pSettingKey, "#name")
        local valueIndex = THUtils.getXMLValue(xmlFile, pSettingKey, "#valueIndex", -1)
        local settingData = self:getSetting(settingName)
        if settingData == nil then
            THUtils.xmlMsg(pSettingKey, "name", settingName, false, THXMLMessage.INVALID_VALUE)
        elseif valueIndex <= 0 then
            THUtils.xmlMsg(pSettingKey, "valueIndex", valueIndex, true, THXMLMessage.INVALID_VALUE)
        else
            if self.isServer or not settingData.syncRequired then
                settingData:setValueByIndex(valueIndex, true)
                settingData.savedValue = settingData.changedValue
            end
        end
    end)
    return loadSuccess
end
function THSettingsManager:saveSettingsToXMLFile(xmlFile, xmlKey)
    local settingsArray, numSettings = self:getSettingsList(true)
    local saveSuccess = true
    if numSettings > 0 then
        THUtils.call(function()
            local settingsBaseKey = string.format("%s.settings", xmlKey)
            local settingXMLIndex = 0
            local areSettingsChanged = false
            xmlFile:removeProperty(settingsBaseKey)
            for settingIndex = 1, numSettings do
                local settingData = settingsArray[settingIndex]
                local settingXMLKey = string.format("%s.setting(%d)", settingsBaseKey, settingXMLIndex)
                if self.isServer or not settingData.syncRequired then
                    local changedValue, valueIndex = settingData:getChangedValue()
                    if changedValue ~= nil and valueIndex > 0 then
                        xmlFile:setString(settingXMLKey .. "#name", settingData.name)
                        xmlFile:setInt(settingXMLKey .. "#valueIndex", valueIndex)
                        if settingData.savedValue ~= valueIndex then
                            settingData.savedValue = valueIndex
                            areSettingsChanged = true
                        end
                        settingXMLIndex = settingXMLIndex + 1
                    end
                end
            end
            if areSettingsChanged then
                self.messageCenter:publish(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, self)
                if self.isServer then
                    self:raiseDirtyFlags(self.dirtyFlag)
                end
            end
        end)
    end
    return saveSuccess
end
function THSettingsManager:loadFromModSettingsXMLFile(xmlFile, xmlKey)
    local loadSuccess = THSettingsManager:superClass().loadFromModSettingsXMLFile(self, xmlFile, xmlKey)
    if loadSuccess then
        if not self.isServer then
            loadSuccess = self:loadSettingsFromXMLFile(xmlFile, xmlKey)
        end
    end
    return loadSuccess
end
function THSettingsManager:saveToModSettingsXMLFile(xmlFile, xmlKey)
    local saveSuccess = THSettingsManager:superClass().saveToModSettingsXMLFile(self, xmlFile, xmlKey)
    if saveSuccess then
        if not self.isServer then
            saveSuccess = self:saveSettingsToXMLFile(xmlFile, xmlKey)
        end
    end
    return saveSuccess
end
function THSettingsManager:loadFromSavegameXMLFile(xmlFile, xmlKey)
    local loadSuccess = THSettingsManager:superClass().loadFromSavegameXMLFile(self, xmlFile, xmlKey)
    if loadSuccess then
        loadSuccess = self:loadSettingsFromXMLFile(xmlFile, xmlKey)
    end
    return loadSuccess
end
function THSettingsManager:saveToSavegameXMLFile(xmlFile, xmlKey)
    local saveSuccess = THSettingsManager:superClass().saveToSavegameXMLFile(self, xmlFile, xmlKey)
    if saveSuccess then
        saveSuccess = self:saveSettingsToXMLFile(xmlFile, xmlKey)
    end
    return saveSuccess
end
function THSettingsManager:onPostInit()
    if not self.isPostInitFinished then
        local rootEnv = self.rootEnvironment
        THSettingsManager:superClass().onPostInit(self)
        self.xmlSchemaSettings = rootEnv.xmlSchemaSettings
        self.xmlSchemaSavegame = rootEnv.xmlSchemaSavegame
        self.isPostInitFinished = true
    end
end
function THSettingsManager:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
    THSettingsManager:superClass().onSetMissionInfo(self, mission, missionInfo, missionDynamicInfo)
    if self.isServer then
        self:loadFromSavegame()
    else
        self:loadModSettings()
    end
    self.areSettingsLoaded = true
end
function THSettingsManager:onUpdateTick(dt)
    THSettingsManager:superClass().onUpdateTick(self, dt)
    if not self.isServer and self.isClientDirty then
        self:saveModSettings()
        self.isClientDirty = false
    end
end
function THSettingsManager:onReadStream(streamId, connection)
    THSettingsManager:superClass().onReadStream(self, streamId, connection)
    if connection:getIsServer() then
        self:readSettingsFromStream(streamId, connection)
    end
end
function THSettingsManager:onWriteStream(streamId, connection)
    THSettingsManager:superClass().onWriteStream(self, streamId, connection)
    if not connection:getIsServer() then
        self:writeSettingsToStream(streamId, connection)
    end
end
function THSettingsManager:onReadUpdateStream(streamId, timeStamp, connection)
    THSettingsManager:superClass().onReadUpdateStream(self, streamId, timeStamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isDirty
            self:readSettingsFromStream(streamId, connection)
        end
    end
end
function THSettingsManager:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(self.dirtyFlag, dirtyMask) ~= 0) then
            self:writeSettingsToStream(streamId, connection)
        end
    end
end
function THSettingsManager:readSettingsFromStream(streamId, connection)
    if connection:getIsServer() then
        local settingsArray, numSettings = self:getSettingsList(true)
        if numSettings > 0 then
            for settingIndex = 1, numSettings do
                local settingData = settingsArray[settingIndex]
                if streamReadBool(streamId) then -- syncRequired
                    settingData.changedValue = streamReadUIntN(streamId, THSettingsManager.SETTING_VALUE_NUM_BITS)
                    settingData.currentValue = streamReadUIntN(streamId, THSettingsManager.SETTING_VALUE_NUM_BITS)
                    settingData.savedValue = streamReadUIntN(streamId, THSettingsManager.SETTING_VALUE_NUM_BITS)
                    settingData.defaultValue = streamReadUIntN(streamId, THSettingsManager.SETTING_VALUE_NUM_BITS)
                end
            end
            self.messageCenter:publish(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, self)
        end
    end
end
function THSettingsManager:writeSettingsToStream(streamId, connection)
    if not connection:getIsServer() then
        local settingsArray, numSettings = self:getSettingsList(true)
        if numSettings > 0 then
            for settingIndex = 1, numSettings do
                local settingData = settingsArray[settingIndex]
                if streamWriteBool(streamId, settingData.syncRequired == true) then
                    streamWriteUIntN(streamId, settingData.changedValue, THSettingsManager.SETTING_VALUE_NUM_BITS)
                    streamWriteUIntN(streamId, settingData.currentValue, THSettingsManager.SETTING_VALUE_NUM_BITS)
                    streamWriteUIntN(streamId, settingData.savedValue, THSettingsManager.SETTING_VALUE_NUM_BITS)
                    streamWriteUIntN(streamId, settingData.defaultValue, THSettingsManager.SETTING_VALUE_NUM_BITS)
                end
            end
        end
    end
end
function THSettingsManager:addSettingCategory(name, title, customEnv, showInMenu)
    customEnv = THUtils.getNoNil(customEnv, self.modName)
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name)
        and THUtils.argIsValid(title == nil or type(title) == "string", "title", title)
        and THUtils.argIsValid(type(customEnv) == "string", "customEnv", customEnv)
    then
        local i18n = self.i18n
        local categoryId = name:upper()
        local constId = THUtils.snakeCase(name, true)
        local categoryData = self.settingCategories.byId[categoryId]
        if categoryData ~= nil then
            THUtils.warningMsg("Setting category (%s) already registered", categoryData.name)
        elseif self.SETTING_CATEGORY[constId] ~= nil then
            THUtils.warningMsg("Setting category constant (%s) already registered", constId)
        else
            local titleText = name
            if title == nil then
                if i18n:hasText("thSettingCategory_" .. name, customEnv) then
                    titleText = i18n:getText("thSettingCategory_" .. name, customEnv)
                end
            else
                titleText = i18n:convertText(title, customEnv)
            end
            categoryData = {
                id = categoryId,
                name = name,
                index = #self.settingCategories.byIndex + 1,
                title = titleText,
                showInMenu = showInMenu ~= false,
                settings = {}
            }
            self.settingCategories.byIndex[categoryData.index] = categoryData
            self.settingCategories.byId[categoryData.id] = categoryData
            self.SETTING_CATEGORY[constId] = categoryData.index
            self.messageCenter:publish(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, self)
        end
        return categoryData
    end
end
function THSettingsManager:addSetting(name, category, title, toolTip, elementType, customEnv, restartRequired, syncRequired, isAdminOnly, showInMenu)
    category = THUtils.getNoNil(category, "main")
    customEnv = THUtils.getNoNil(customEnv, self.modName)
    local settingData = nil
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name)
        and THUtils.argIsValid(title == nil or type(title) == "string", "title", title)
        and THUtils.argIsValid(toolTip == nil or type(toolTip) == "string", "toolTip", toolTip)
        and THUtils.argIsValid(type(customEnv) == "string", "customEnv", customEnv)
    then
        local i18n = self.i18n
        local settingId = name:upper()
        local constId = THUtils.snakeCase(name, true)
        local categoryData = self:getSettingCategory(category)
        if categoryData == nil then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "category", category)
        elseif elementType == nil or THSettingsManager.ELEMENT_TYPE_ID[elementType] == nil then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "elementType", elementType)
        else
            settingData = self.settings.byId[settingId]
            if settingData ~= nil then
                THUtils.warningMsg("Setting (%s) already registered", settingData.name)
            elseif self.SETTING[constId] ~= nil then
                THUtils.warningMsg("Setting constant (%s) already registered", constId)
            else
                local titleText = name
                local toolTipText = ""
                if title == nil then
                    if i18n:hasText("thSetting_" .. name, customEnv) then
                        titleText = i18n:getText("thSetting_" .. name, customEnv)
                    end
                else
                    titleText = i18n:convertText(title, customEnv)
                end
                if toolTip == nil then
                    if i18n:hasText("thSettingToolTip_" .. name, customEnv) then
                        toolTipText = i18n:getText("thSettingToolTip_" .. name, customEnv)
                    end
                else
                    toolTipText = i18n:convertText(toolTip, customEnv)
                end
                settingData = {
                    id = settingId,
                    name = name,
                    index = #self.settings.byIndex + 1,
                    manager = self,
                    customEnv = customEnv,
                    title = titleText,
                    toolTip = toolTipText,
                    elementType = elementType,
                    categoryIndex = categoryData.index,
                    isClient = self.isClient,
                    isServer = self.isServer,
                    isAdminOnly = syncRequired == true and isAdminOnly == true,
                    syncRequired = syncRequired == true,
                    restartRequired = restartRequired == true,
                    showInMenu = showInMenu ~= false,
                    isLocked = false,
                    values = {},
                    valueMapping = {},
                    valueTexts = {},
                    currentValue = 0,
                    changedValue = 0,
                    savedValue = 0,
                    defaultValue = 0
                }
                setmetatable(settingData, THSetting_mt)
                if settingData.elementType == THSettingsManager.ELEMENT_TYPE.CHECKED then
                    settingData:addValue(false, "$l10n_ui_off")
                    settingData:addValue(true, "$l10n_ui_on")
                    settingData.isLocked = true
                end
                settingData.messageType = nextMessageTypeId()
                self.settings.byId[settingData.id] = settingData
                self.settings.byIndex[settingData.index] = settingData
                self.SETTING[constId] = settingData.index
                table.insert(categoryData.settings, settingData.index)
                self.messageCenter:publish(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, self)
            end
        end
    end
    return settingData
end
function THSettingsManager:addSettingChangedListener(setting, target, callbackFunc)
    local settingData = self:getSetting(setting)
    if settingData ~= nil
        and THUtils.argIsValid(type(target) == "table", "target", target)
        and THUtils.argIsValid(type(callbackFunc) == "function", "callbackFunc", callbackFunc)
    then
        self.messageCenter:unsubscribe(settingData.messageType, target)
        self.messageCenter:subscribe(settingData.messageType, callbackFunc, target)
    end
end
function THSettingsManager:removeSettingChangedListener(setting, target)
    local settingData = self:getSetting(setting)
    if settingData ~= nil then
        self.messageCenter:unsubscribe(settingData.messageType, target)
    end
end
function THSettingsManager:getSettingCategory(category)
    local valType = type(category)
    local categoryData = nil
    if valType == "string" then
        local categoryId = category:upper()
        categoryData = self.settingCategories.byId[categoryId]
        if categoryData == nil then
            local categoryIndex = self.SETTING_CATEGORY[categoryId]
            if categoryIndex ~= nil then
                categoryData = self.settingCategories.byIndex[categoryIndex]
            end
        end
    elseif valType == "number" then
        categoryData = self.settingCategories.byIndex[category]
    elseif valType == "table" and category.id ~= nil then
        categoryData = self.settingCategories.byId[category.id]
        if categoryData ~= category then
            categoryData = nil
        end
    elseif category ~= nil then -- nil values pass though
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "category", category)
    end
    return categoryData
end
function THSettingsManager:getSettingCategoryList(byIndex)
    local numCategories = #self.settingCategories.byIndex
    if byIndex == true then
        return self.settingCategories.byIndex, numCategories
    else
        return self.settingCategories.byId, numCategories
    end
end
function THSettingsManager:getSetting(setting)
    local valType = type(setting)
    local settingData = nil
    if valType == "string" then
        local settingId = setting:upper()
        settingData = self.settings.byId[settingId]
        if settingData == nil then
            local settingIndex = self.SETTING[settingId]
            if settingIndex ~= nil then
                settingData = self.settings.byIndex[settingIndex]
            end
        end
    elseif valType == "number" then
        settingData = self.settings.byIndex[setting]
    elseif valType == "table" and setting.id ~= nil then
        settingData = self.settings.byId[setting.id]
        if settingData ~= setting then
            settingData = nil
        end
    elseif setting ~= nil then -- nil values pass through
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "setting", setting)
    end
    return settingData
end
function THSettingsManager:getSettingsList(byIndex)
    local numSettings = #self.settings.byIndex
    if byIndex == true then
        return self.settings.byIndex, numSettings
    else
        return self.settings.byId, numSettings
    end
end
function THSettingsManager:getSettingValue(setting, verbose)
    local settingData = self:getSetting(setting)
    if settingData == nil then
        THUtils.msgOnTrue(verbose == true, THMessage.ARGUMENT_INVALID, "setting", setting)
    else
        local settingValue, valueIndex = settingData:getCurrentValue()
        return settingValue, valueIndex, settingData
    end
end
function THSettingsManager:setSettingValue(setting, value, noEventSend, updateDefault)
    local settingData = self:getSetting(setting)
    if settingData ~= nil then
        return settingData:setValue(value, noEventSend, updateDefault)
    end
end
function THSetting:getCanAccess()
    if self.isAdminOnly then
        local isMasterUser = THUtils.getIsMasterUser()
        if not isMasterUser then
            return false
        end
    end
    return true
end
function THSetting:getCurrentValue()
    return self.values[self.currentValue], self.currentValue
end
function THSetting:getChangedValue()
    return self.values[self.changedValue], self.changedValue
end
function THSetting:getDefaultValue()
    return self.values[self.defaultValue], self.defaultValue
end
function THSetting:addValue(value, title)
    if THUtils.argIsValid(value ~= nil, "value", value)
        and THUtils.argIsValid(type(title) == "string" and title ~= "", "title", title)
    then
        local manager = self.manager
        local i18n = manager.i18n
        local valType = type(value)
        local valueId = value
        if valType == "string" then
            valueId = value:upper()
        end
        if self.valueMapping[valueId] ~= nil then
            THUtils.warningMsg("Value (%s) has already been added to setting (%s)", value, self.name)
        else
            local valueIndex = #self.values + 1
            local isSettingDirty = false
            self.values[valueIndex] = value
            self.valueTexts[valueIndex] = i18n:convertText(title, self.customEnv)
            self.valueMapping[valueId] = valueIndex
            if self.changedValue <= 0 then
                self.changedValue = 1
                isSettingDirty = true
            end
            if self.currentValue <= 0 then
                self.currentValue = 1
                isSettingDirty = true
            end
            if self.defaultValue <= 0 then
                self.defaultValue = 1
                isSettingDirty = true
            end
            if self.savedValue <= 0 then
                self.savedValue = 1
                isSettingDirty = true
            end
            if isSettingDirty then
                manager.messageCenter:publish(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, manager)
                if manager.isServer then
                    manager:raiseDirtyFlags(manager.dirtyFlag)
                end
            end
        end
    end
end
function THSetting:setValue(value, noEventSend, updateDefault)
    local valType = type(value)
    local valueIndex = -1
    if valType == "string" then
        valueIndex = self.valueMapping[value:upper()]
    elseif value ~= nil then
        valueIndex = self.valueMapping[value]
    else
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "value", value)
    end
    if valueIndex ~= nil and valueIndex > 0 then
        local newValue, newValueIndex = self:setValueByIndex(valueIndex, noEventSend, updateDefault)
        if newValue ~= nil then
            return newValueIndex, newValue
        end
    end
    return valueIndex
end
function THSetting:setValueByIndex(valueIndex, noEventSend, updateDefault)
    updateDefault = THUtils.getNoNil(updateDefault, false)
    if THUtils.argIsValid(type(valueIndex) == "number", "valueIndex", valueIndex)
        and THUtils.argIsValid(type(updateDefault) == "boolean", "updateDefault", updateDefault)
    then
        local manager = self.manager
        local isMasterUser = THUtils.getIsMasterUser()
        local newValue = self.values[valueIndex]
        if newValue ~= nil then
            if self.isServer or not self.syncRequired or noEventSend == true then
                self.changedValue = valueIndex
                local isCurrentValueChanged = false
                local isSettingDirty = false
                if not manager.areSettingsLoaded or not self.restartRequired then
                    if valueIndex ~= self.currentValue then
                        self.currentValue = valueIndex
                        isCurrentValueChanged = true
                        isSettingDirty = true
                    end
                    if not manager.areSettingsLoaded then
                        self.savedValue = valueIndex
                    end
                end
                if updateDefault == true then
                    if self.defaultValue ~= valueIndex then
                        self.defaultValue = valueIndex
                        isSettingDirty = true
                    end
                end
                if isSettingDirty then
                    if isCurrentValueChanged then
                        manager.messageCenter:publish(self.messageType, self, self.values[self.currentValue])
                    end
                    manager.messageCenter:publish(THSettingsManager.MESSAGE_TYPE.SETTINGS_CHANGED, manager)
                    if noEventSend ~= true then
                        if self.isServer then
                            manager:raiseDirtyFlags(manager.dirtyFlag)
                        elseif not self.syncRequired then
                            manager.isClientDirty = true
                            manager:raiseActive()
                        end
                    end
                end
                return newValue, valueIndex
            elseif isMasterUser or not self.isAdminOnly then
                if THSetSettingValueEvent.sendEvent(manager, self.index, valueIndex, updateDefault) then
                    return newValue, valueIndex
                end
            end
        end
    end
    return nil, -1
end
THSetSettingValueEvent = {}
local THSetSettingValueEvent_mt = THUtils.createClass(THSetSettingValueEvent, Event)
InitEventClass(THSetSettingValueEvent, "THSetSettingValueEvent")
function THSetSettingValueEvent.emptyNew()
    return Event.new(THSetSettingValueEvent_mt)
end
function THSetSettingValueEvent.new(manager, settingIndex, valueIndex, updateDefault)
    local self = THSetSettingValueEvent.emptyNew()
    self.thIsEventValid = false
    if THUtils.argIsValid(THUtils.getIsType(manager, THSettingsManager), "manager", manager)
        and THUtils.argIsValid(type(settingIndex) == "number" and settingIndex > 0, "settingIndex", settingIndex)
        and THUtils.argIsValid(type(valueIndex) == "number" and valueIndex > 0, "valueIndex", valueIndex)
        and THUtils.argIsValid(type(updateDefault) == "boolean", "updateDefault", updateDefault)
    then
        local networkId = manager.networkId or 0
        if networkId <= 0 then
            THUtils.stackMsg("THSettingsManager: invalid network id (%s)", networkId)
        else
            self.networkId = networkId
            self.settingIndex = settingIndex
            self.valueIndex = valueIndex
            self.updateDefault = updateDefault
            self.thIsEventValid = true
        end
    end
    return self
end
function THSetSettingValueEvent:readStream(streamId, connection)
    if not connection:getIsServer() then -- server reading from client
        if streamReadBool(streamId) then -- is event valid
            self.networkId = streamReadUIntN(streamId, THAbstractSystem.NETWORK_ID_BITS)
            self.settingIndex = streamReadUIntN(streamId, THSettingsManager.SETTING_NUM_BITS)
            self.valueIndex = streamReadUIntN(streamId, THSettingsManager.SETTING_VALUE_NUM_BITS)
            self.updateDefault = streamReadBool(streamId)
            self:run(connection)
        end
    else
        THUtils.stackMsg("THSetSettingValueEvent should not be sent from the server!")
    end
end
function THSetSettingValueEvent:writeStream(streamId, connection)
    if connection:getIsServer() then -- client sending to server
        if streamWriteBool(streamId, self.thIsEventValid) then
            streamWriteUIntN(streamId, self.networkId, THAbstractSystem.NETWORK_ID_BITS)
            streamWriteUIntN(streamId, self.settingIndex, THSettingsManager.SETTING_NUM_BITS)
            streamWriteUIntN(streamId, self.valueIndex, THSettingsManager.SETTING_VALUE_NUM_BITS)
            streamWriteBool(streamId, self.updateDefault)
        end
    else
        THUtils.stackMsg("THSetSettingValueEvent: writeStream should not run on the server!")
    end
end
function THSetSettingValueEvent:run(connection)
    if not connection:getIsServer() then -- server reading from client
        local thSettingsManager = THAbstractSystem.getTargetByNetworkId(self.networkId)
        if not THUtils.getIsType(thSettingsManager, THSettingsManager) then
            THUtils.warningMsg("THSettingsManager: invalid network id (%s)", self.networkId)
        else
            local settingData = thSettingsManager:getSetting(self.settingIndex)
            if settingData == nil then
                THUtils.stackMsg("Invalid setting index (%s) sent to server!", self.settingIndex)
            elseif not settingData.syncRequired then
                THUtils.stackMsg("Non multiplayer setting (%s) should not be synchronized!", self.settingIndex)
            else
                settingData:setValueByIndex(self.valueIndex, nil, self.updateDefault)
            end
        end
    end
end
function THSetSettingValueEvent.sendEvent(manager, settingIndex, valueIndex, updateDefault)
    if g_server == nil and g_client ~= nil then
        local newEvent = THSetSettingValueEvent.new(manager, settingIndex, valueIndex, updateDefault)
        if newEvent ~= nil and newEvent.thIsEventValid then
            g_client:getServerConnection():sendEvent(newEvent)
            return true
        end
    end
    return false
end