-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THDebugUtil = {}
THDebugUtil.FLAG_UNKNOWN = "UNKNOWN"
THDebugLevel = {
    ERROR = 1,
    WARNING = 2,
    NORMAL = 3,
    UPDATE = 4,
    LOOP = 5
}
local thDebugFlags = {}
local thDebugLevel = THDebugLevel.NORMAL
local thGlobalDebugFlag = false
function THDebugUtil.printMsg(flagId, debugLevel, text, ...)
    local realFlagId = THDebugUtil.getDebugFlag(flagId)
    if THDebugUtil.getIsEnabled(realFlagId, debugLevel) then
        if text == "" then
            printf("")
        else
            if realFlagId ~= nil and realFlagId ~= "" then
                text = string.format("[%s] %s", realFlagId, text)
            end
            THUtils.displayMsg("DEBUG: " .. text, ...)
        end
    end
end
function THDebugUtil.printTable(flagId, debugLevel, tbl, ...)
    if THDebugUtil.getIsEnabled(flagId, debugLevel) then
        THUtils.printTable(tbl, ...)
    end
end
function THDebugUtil.createDebugFlag(flagName)
    if type(flagName) == "string" and flagName ~= "" then
        local flagId = flagName:upper() -- case insensitive
        if thDebugFlags[flagId] == nil then
            thDebugFlags[flagId] = false
        end
        return flagId
    else
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "flagName", flagName)
    end
end
function THDebugUtil.getDebugFlag(flagId)
    if flagId == nil or flagId == "" then
        return ""
    end
    local flagIdType = type(flagId)
    local realFlagId = nil
    if flagIdType == "string" then
        realFlagId = flagId
    elseif flagIdType == "table" then
        realFlagId = flagId.debugFlagId
    end
    if type(realFlagId) == "string" then
        realFlagId = realFlagId:upper()
        if thDebugFlags[realFlagId] ~= nil then
            return realFlagId
        end
    end
    return THDebugUtil.FLAG_UNKNOWN
end
function THDebugUtil.getDebugFlagList()
    return thDebugFlags
end
function THDebugUtil.getIsEnabled(flagId, reqDebugLevelId)
    local reqDebugLevel = nil
    if reqDebugLevelId == nil then
        reqDebugLevel = thDebugLevel
    else
        reqDebugLevel = THDebugUtil.getDebugLevelById(reqDebugLevelId)
    end
    local realFlagId = THDebugUtil.getDebugFlag(flagId)
    local isDebugEnabled = false
    if realFlagId == nil or realFlagId == THDebugUtil.FLAG_UNKNOWN then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "flagId", flagId)
    elseif reqDebugLevel ~= nil and thDebugLevel >= reqDebugLevel then
        if realFlagId == "" then
            isDebugEnabled = thGlobalDebugFlag == true
        else
            isDebugEnabled = thDebugFlags[realFlagId] == true
        end
    end
    return isDebugEnabled
end
function THDebugUtil.setIsEnabled(flagId, isEnabled)
    local realFlagId = THDebugUtil.getDebugFlag(flagId)
    if realFlagId == "" then
        if isEnabled == true then
            thGlobalDebugFlag = true
        else
            thGlobalDebugFlag = false
        end
        for otherFlagId in pairs(thDebugFlags) do
            thDebugFlags[otherFlagId] = isEnabled
        end
        return thGlobalDebugFlag == true
    end
    if realFlagId ~= nil and realFlagId ~= THDebugUtil.FLAG_UNKNOWN
        and thDebugFlags[realFlagId] ~= nil
    then
        if isEnabled == true then
            thDebugFlags[realFlagId] = true
        else
            thDebugFlags[realFlagId] = false
        end
        return thDebugFlags[realFlagId] == true
    end
    THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "flagId", flagId)
    return false
end
function THDebugUtil.getDebugLevel()
    return thDebugLevel or THDebugLevel.NORMAL
end
function THDebugUtil.getDebugLevelById(debugLevelId)
    local debugLevel = nil -- placeholder
    if type(debugLevelId) == "string" then
        debugLevel = THDebugLevel[debugLevelId:upper()]
    elseif debugLevelId ~= nil then
        debugLevel = tonumber(debugLevelId)
    end
    if debugLevel == nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "debugLevelId", debugLevelId)
    end
    return debugLevel
end
function THDebugUtil.setDebugLevel(debugLevelId)
    local debugLevel = THDebugUtil.getDebugLevelById(debugLevelId)
    if debugLevel ~= nil then
        thDebugLevel = debugLevel
    end
end
function THDebugUtil.reset()
    THDebugUtil.setIsEnabled(nil, false)
    THDebugUtil.setDebugLevel(THDebugLevel.NORMAL)
end
THSetIsDebugEnabledEvent = {}
local THSetIsDebugEnabledEvent_mt = THUtils.createClass(THSetIsDebugEnabledEvent, Event)
InitEventClass(THSetIsDebugEnabledEvent, "THSetIsDebugEnabledEvent")
function THSetIsDebugEnabledEvent.emptyNew()
    return Event.new(THSetIsDebugEnabledEvent_mt)
end
function THSetIsDebugEnabledEvent.new(flagId, isEnabled)
    local self = THSetIsDebugEnabledEvent.emptyNew()
    self.thIsEventValid = false
    if THUtils.argIsValid(type(flagId) == "string", "flagId", flagId)
        and THUtils.argIsValid(type(isEnabled) == "boolean", "isEnabled", isEnabled)
    then
        self.flagId = flagId
        self.isEnabled = isEnabled
        self.thIsEventValid = true
    end
    return self
end
function THSetIsDebugEnabledEvent:readStream(streamId, connection)
    if not connection:getIsServer() then -- server reading from client
        if streamReadBool(streamId) then -- is event valid
            self.flagId = streamReadString(streamId)
            self.isEnabled = streamReadBool(streamId)
            self:run(connection)
        end
    else
        THUtils.stackMsg("THSetIsDebugEnabledEvent should not read from clients!")
    end
end
function THSetIsDebugEnabledEvent:writeStream(streamId, connection)
    if connection:getIsServer() then -- client sending to server
        if streamWriteBool(streamId, self.thIsEventValid) then
            streamWriteString(streamId, self.flagId)
            streamWriteBool(streamId, self.isEnabled)
        end
    else
        THUtils.stackMsg("THSetIsDebugEnabledEvent should not write from server!")
    end
end
function THSetIsDebugEnabledEvent:run(connection)
    if not connection:getIsServer() then
        THDebugUtil.setIsEnabled(self.flagId, self.isEnabled)
    end
end
function THSetIsDebugEnabledEvent.sendEvent(flagId, isEnabled)
    THUtils.call(function()
        flagId = flagId or ""
        local mission = g_thAdvancedAddon.mission
        if mission ~= nil and g_server == nil and g_client ~= nil then
            if mission.isMasterUser then
                local newEvent = THSetIsDebugEnabledEvent.new(flagId, isEnabled)
                if newEvent ~= nil and newEvent.thIsEventValid then
                    g_client:getServerConnection():sendEvent(newEvent)
                end
            else
                THUtils.displayMsg(THMessage.COMMAND_REQUIRES_MASTER_USER)
            end
        end
    end)
end
THSetDebugLevelEvent = {}
local THSetDebugLevelEvent_mt = THUtils.createClass(THSetDebugLevelEvent, Event)
InitEventClass(THSetDebugLevelEvent, "THSetDebugLevelEvent")
function THSetDebugLevelEvent.emptyNew()
    return Event.new(THSetDebugLevelEvent_mt)
end
function THSetDebugLevelEvent.new(debugLevel)
    local self = THSetDebugLevelEvent.emptyNew()
    self.thIsEventValid = false
    if THUtils.argIsValid(type(debugLevel) == "number", "debugLevel", debugLevel) then
        self.debugLevel = debugLevel
        self.thIsEventValid = true
    end
    return self
end
function THSetDebugLevelEvent:readStream(streamId, connection)
    if not connection:getIsServer() then -- server reading from client
        if streamReadBool(streamId) then -- is event valid
            self.debugLevel = streamReadFloat32(streamId)
            self:run(connection)
        end
    else
        THUtils.stackMsg("THSetDebugLevelEvent should not read from client!")
    end
end
function THSetDebugLevelEvent:writeStream(streamId, connection)
    if connection:getIsServer() then -- client sending to server
        if streamWriteBool(streamId, self.thIsEventValid) then
            streamWriteFloat32(streamId, self.debugLevel)
        end
    else
        THUtils.stackMsg("THSetDebugLevelEvent should not write from server!")
    end
end
function THSetDebugLevelEvent:run(connection)
    if not connection:getIsServer() then -- server reading from client
        THDebugUtil.setDebugLevel(self.debugLevel)
    end
end
function THSetDebugLevelEvent.sendEvent(debugLevel)
    THUtils.call(function()
        local mission = g_thAdvancedAddon.mission
        if mission ~= nil and g_server == nil and g_client ~= nil then
            if mission.isMasterUser then
                local newEvent = THSetDebugLevelEvent.new(debugLevel)
                if newEvent ~= nil and newEvent.thIsEventValid then
                    g_client:getServerConnection():sendEvent(newEvent)
                end
            else
                THUtils.displayMsg(THMessage.COMMAND_REQUIRES_MASTER_USER)
            end
        end
    end)
end