-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thInputManager = g_thAdvancedAddon.thInputManager
THVehicleInfoDisplay = {}
local THVehicleInfoDisplay_mt = THUtils.createClass(THVehicleInfoDisplay, InfoDisplay)
THVehicleInfoDisplay.OFFSET = {
    BACKGROUND = { 0, 40 }
}
local function getCustomData(self)
    local customData = nil
    if self ~= nil then
        customData = rawget(self, self)
    end
    return customData
end
THVehicleInfoDisplay.getCustomData = getCustomData
function THVehicleInfoDisplay.new(hudAtlasPath, uiScale, customMt)
    customMt = customMt or THVehicleInfoDisplay_mt
    local backgroundOverlay = THVehicleInfoDisplay.createBackground(hudAtlasPath)
    local self = InfoDisplay:superClass().new(backgroundOverlay, nil, customMt)
    self.isClient = g_thAdvancedAddon.isClient
    self.isServer = g_thAdvancedAddon.isServer
    self.boxes = {}
    self.boxMarginY = 0
    self.isEnabled = true
    local customData = {}
    customData.offsetX = 0
    customData.offsetY = 0
    customData.isHudActive = true
    customData.actionEventIds = {}
    rawset(self, self, customData)
    self:setScale(uiScale)
    self:storeOriginalPosition()
    self:setVisible(false, false)
    return self
end
function THVehicleInfoDisplay:update(dt, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if self.isClient and customData ~= nil then
            THUtils.call(function()
                self:updateActionEvents(dt)
            end)
        end
        return ...
    end
    return appendFunc(THVehicleInfoDisplay:superClass().update(self, dt, ...))
end
function THVehicleInfoDisplay:draw(...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if self.isEnabled and customData ~= nil and customData.isHudActive then
            THUtils.call(function()
                local posX = customData.offsetX
                local posY = customData.offsetY
                for elementIndex = #self.boxes, 1, -1 do
                    local boxElement = self.boxes[elementIndex]
                    if boxElement:canDraw() then
                        boxElement:draw(posX, posY)
                        posY = posY + boxElement:getDisplayHeight() + self.boxMarginY
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(InfoDisplay:superClass().draw(self, ...))
end
function THVehicleInfoDisplay:onActionToggleVisible()
    local customData = getCustomData(self)
    if self.isClient and customData ~= nil then
        customData.isHudActive = not customData.isHudActive
    end
end
function THVehicleInfoDisplay:setVehicle(vehicle)
    local customData = getCustomData(self)
    if customData ~= nil then
        customData.vehicle = vehicle
        self:setEnabled(customData.vehicle ~= nil)
    end
end
function THVehicleInfoDisplay:registerInput()
    local customData = getCustomData(self)
    if self.isClient and customData ~= nil then
        local actionEventIds = customData.actionEventIds
        self:unregisterInput()
        local _, actionEventId = thInputManager:registerActionEvent("THAA_VEHICLE_FUNCTION", 1, self, THVehicleInfoDisplay.onActionToggleVisible, false, true, false, true)
        if actionEventId ~= nil then
            actionEventIds.toggleVisible = actionEventId
            thInputManager:setActionEventTextVisibility(actionEventId, false)
        end
    end
end
function THVehicleInfoDisplay:unregisterInput()
    local customData = getCustomData(self)
    if self.isClient and customData ~= nil then
        for _, actionEventId in pairs(customData.actionEventIds) do
            thInputManager:removeActionEvent(actionEventId)
        end
    end
end
function THVehicleInfoDisplay:updateActionEvents(dt)
    local customData = getCustomData(self)
    if self.isClient and customData ~= nil then
        local actionEventIds = customData.actionEventIds
        local vehicle = customData.vehicle
        if actionEventIds.toggleVisible ~= nil then
            local canToggleVisible = false
            if self.isEnabled then
                if vehicle ~= nil and vehicle.getIsActiveForInput ~= nil and vehicle:getIsActiveForInput(true, true) then
                    canToggleVisible = true
                end
            end
            thInputManager:setActionEventActive(actionEventIds.toggleVisible, canToggleVisible)
        end
    end
end
function THVehicleInfoDisplay.getBackgroundPosition(uiScale)
    local width = getNormalizedScreenValues(THUtils.unpack(InfoDisplay.SIZE.SELF))
    local _, height = getNormalizedScreenValues(THUtils.unpack(SpeedMeterDisplay.SIZE.BACKGROUND))
    local posX = 1 - g_safeFrameOffsetX - (width * uiScale)
    local posY = g_safeFrameOffsetY + (height * uiScale)
    return posX, posY
end
function THVehicleInfoDisplay.createBackground(hudAtlasPath, ...)
    local function appendFunc(rOverlay, ...)
        if rOverlay ~= nil then
            THUtils.call(function()
                local posX, posY = THVehicleInfoDisplay.getBackgroundPosition(1)
                rOverlay:setPosition(posX, posY)
            end)
        end
        return rOverlay, ...
    end
    return appendFunc(THVehicleInfoDisplay:superClass().createBackground(hudAtlasPath, ...))
end
function THVehicleInfoDisplay:storeScaledValues(...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                local _, height = self:scalePixelToScreenVector(SpeedMeterDisplay.SIZE.BACKGROUND)
                local _, offsetY = self:scalePixelToScreenVector(THVehicleInfoDisplay.OFFSET.BACKGROUND)
                customData.offsetX = 1 - g_safeFrameOffsetX
                customData.offsetY = g_safeFrameOffsetY + height + offsetY
            end)
        end
        return ...
    end
    return appendFunc(THVehicleInfoDisplay:superClass().storeScaledValues(self, ...))
end