-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thAdvancedAddon.modName
local thFillTypeManager = g_thAdvancedAddon.thFillTypeManager
local thFruitTypeManager = g_thAdvancedAddon.thFruitTypeManager
THCombine = {}
local debugFlagId = THDebugUtil.createDebugFlag("THCombine")
THCombine.debugFlagId = debugFlagId
THCombine.MOD_NAME = thModName
THCombine.SPEC_NAME = "combine"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THCombine.SPEC_NAME, thModName, verbose)
end
THCombine.getSpecTable = getSpecTable
function THCombine.initSpecialization()
end
function THCombine.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THVehicle, specializations)
        and SpecializationUtil.hasSpecialization(Combine, specializations)
end
function THCombine.registerEvents(vehicleType)
end
function THCombine.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THCombine)
end
function THCombine.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getCombineFillLevel", THCombine.thAA_getCombineFillLevel)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getCombineSupportedFillTypes", THCombine.thAA_getCombineSupportedFillTypes)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getCombineLastValidCutter", THCombine.thAA_getCombineLastValidCutter)
end
function THCombine.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "addCutterArea", THCombine.addCutterArea)
end
function THCombine:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.lastCutterParameters = {}
    specTable.combineFillTypes = {}
    specTable.bufferFillTypes = {}
end
function THCombine:thAA_getCombineFillLevel(fillTypeIndex, isStrict)
    local combineSpec = self.spec_combine
    if fillTypeIndex == nil and isStrict ~= true then
        fillTypeIndex = self:getCombineLastValidFillType()
    end
    if fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN then
        local function processFillUnit(pFillUnitIndex)
            if pFillUnitIndex ~= nil then
                local combineFillLevel = self:getFillUnitFillLevel(pFillUnitIndex) or 0
                if combineFillLevel > 0 then
                    local combineFillType = self:getFillUnitFillType(pFillUnitIndex)
                    if combineFillType == fillTypeIndex then
                        return combineFillLevel, combineFillType, pFillUnitIndex
                    end
                end
            end
        end
        local combineFillLevel, combineFillType, combineFillUnit = processFillUnit(combineSpec.bufferFillUnitIndex)
        if combineFillLevel == nil then
            combineFillLevel, combineFillType, combineFillUnit = processFillUnit(combineSpec.fillUnitIndex)
        end
        combineFillLevel = combineFillLevel or 0
        combineFillType = combineFillType or FillType.UNKNOWN
        return combineFillLevel, combineFillType, combineFillUnit
    end
    return 0, FillType.UNKNOWN
end
function THCombine:thAA_getCombineSupportedFillTypes(isCombined, refresh)
    local specTable = getSpecTable(self)
    local combineSpec = self.spec_combine
    local hasBufferFillUnit = combineSpec.bufferFillUnitIndex ~= nil
        and combineSpec.bufferFillUnitIndex ~= combineSpec.fillUnitIndex
    if refresh == true or next(specTable.combineFillTypes) == nil then
        THUtils.clearTable(specTable.combineFillTypes)
        THUtils.clearTable(specTable.bufferFillTypes)
        local fillTypeList = nil
        if hasBufferFillUnit then
            fillTypeList = self:getFillUnitSupportedFillTypes(combineSpec.bufferFillUnitIndex)
            if fillTypeList ~= nil then
                for fillTypeIndex in pairs(fillTypeList) do
                    specTable.bufferFillTypes[fillTypeIndex] = true
                    specTable.combineFillTypes[fillTypeIndex] = true
                end
            end
        end
        if combineSpec.fillUnitIndex ~= nil then
            fillTypeList = self:getFillUnitSupportedFillTypes(combineSpec.fillUnitIndex)
            if fillTypeList ~= nil then
                for fillTypeIndex in pairs(fillTypeList) do
                    specTable.combineFillTypes[fillTypeIndex] = true
                end
            end
        end
    end
    if hasBufferFillUnit and isCombined == false then
        return specTable.bufferFillTypes
    end
    return specTable.combineFillTypes
end
function THCombine:thAA_getCombineLastValidCutter()
    local specTable = getSpecTable(self)
    local combineSpec = self.spec_combine
    if specTable.lastValidCutterVehicle ~= nil then
        if combineSpec.attachedCutters ~= nil
            and combineSpec.attachedCutters[specTable.lastValidCutterVehicle] ~= nil
        then
            return specTable.lastValidCutterVehicle
        end
    end
end
function THCombine:addCutterArea(superFunc, realArea, changedArea, inputFruitType, outputFillType, strawRatio, chopperType, farmId, cutterLoad, ...)
    local specTable = getSpecTable(self)
    local cutterParams = specTable.lastCutterParameters
    THUtils.call(function()
        THUtils.clearTable(cutterParams)
        local cutterVehicle = self:thAA_getCombineLastValidCutter()
        local thCutterSpec = THCutter.getSpecTable(cutterVehicle, false)
        if thCutterSpec ~= nil then
            local cutterSpec = cutterVehicle.spec_cutter
            cutterParams.lastStrawArea = thCutterSpec.lastStrawArea or changedArea
            cutterParams.growthState = cutterSpec.currentGrowthState
            cutterParams.farmlandId = thCutterSpec.lastTouchedFarmland
            cutterParams.soilTypeIndex = thCutterSpec.lastTouchedSoilType
        end
    end)
    local function appendFunc(rAppliedDelta, ...)
        if rAppliedDelta ~= nil and rAppliedDelta > 0
            and inputFruitType ~= nil and inputFruitType ~= FruitType.UNKNOWN
            and outputFillType ~= nil and outputFillType ~= FillType.UNKNOWN
        then
            THUtils.call(function()
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local inputFruitTypeName = thFruitTypeManager:getFruitTypeName(inputFruitType)
                    local outputFillTypeName = thFillTypeManager:getFillTypeName(outputFillType)
                    THVehicle.displayMsg(self, "%s >> Last cutter parameters:")
                    THVehicle.displayMsg(self, "- Input fruitType: [%s] (%s)", inputFruitType, inputFruitTypeName)
                    THVehicle.displayMsg(self, "- Output fillType: [%s] (%s)", outputFillType, outputFillTypeName)
                    THVehicle.displayMsg(self, "- Changed Area: %s", changedArea)
                    THVehicle.displayMsg(self, "- Real Area: %s", realArea)
                    THVehicle.displayMsg(self, "- Straw ratio: %s", strawRatio)
                    THVehicle.displayMsg(self, "- Cutter load: %s", cutterLoad)
                    THUtils.printTable(cutterParams)
                end
            end)
        end
        return rAppliedDelta, ...
    end
    return appendFunc(superFunc(self, realArea, changedArea, inputFruitType, outputFillType, strawRatio, chopperType, farmId, cutterLoad, ...))
end